# third party
from nacl.signing import SigningKey
from nacl.signing import VerifyKey

# syft absolute
import syft as sy
from syft import ReprMessage
from syft import serialize
from syft.core.common.message import SignedImmediateSyftMessageWithoutReply
from syft.util import get_fully_qualified_name


def get_signing_key() -> SigningKey:
    # return a the signing key used to sign the get_signed_message_bytes fixture
    key = "e89ff2e651b42393b6ecb5956419088781309d953d72bd73a0968525a3a6a951"
    return SigningKey(bytes.fromhex(key))


def get_signed_message_bytes() -> bytes:
    # return a signed message fixture containing the uid from get_uid
    blob = (
        b"\n?syft.core.common.message.SignedImmediateSyftMessageWithoutReply"
        + b"\x12\xad\x02\n\x12\n\x10\x8c3\x19,\xcd\xd3\xf3N\xe2\xb0\xc6\tU\xdf\x02u\x12"
        + b"6syft.core.node.common.service.repr_service.ReprMessage"
        + b"\x1a@@\x82\x13\xfaC\xfb=\x01H\x853\x1e\xceE+\xc6\xb5\rX\x16Z\xb8l\x02\x10"
        + b"\x8algj\xd6U\x11]\xe9R\x0ei\xd8\xca\xb9\x00=\xa1\xeeoEa\xe2C\xa0\x960\xf7A"
        + b'\xfad<(9\xe1\x8c\x93\xf1\x0b" \x81\xff\xcc\xfc7\xc4U.\x8a*\x1f"=0\x10\xc4'
        + b"\xef\x88\xc80\x01\xf0}3\x0b\xd4\x97\xad/P\x8f\x0f*{\n6"
        + b"syft.core.node.common.service.repr_service.ReprMessage\x12A\n\x12\n\x10"
        + b"\x8c3\x19,\xcd\xd3\xf3N\xe2\xb0\xc6\tU\xdf\x02u\x12+\n\x0bGoofy KirchH\x01R"
        + b"\x1a\n\x12\n\x10\xfb\x1b\xb0g[\xb7LI\xbe\xce\xe7\x00\xab\n\x15\x14\x12\x04Test"
    )
    return blob


def get_repr_message_bytes() -> bytes:
    blob = (
        b"\n6syft.core.node.common.service.repr_service.ReprMessage\x12A\n\x12\n\x10"
        + b"\x8c3\x19,\xcd\xd3\xf3N\xe2\xb0\xc6\tU\xdf\x02u\x12+\n\x0bGoofy KirchH\x01R"
        + b"\x1a\n\x12\n\x10\xfb\x1b\xb0g[\xb7LI\xbe\xce\xe7\x00\xab\n\x15\x14\x12\x04Test"
    )
    return blob


def get_repr_message() -> ReprMessage:
    # return a repr message fixture
    blob = get_repr_message_bytes()
    return sy.deserialize(blob=blob, from_bytes=True)


def get_verify_key() -> VerifyKey:
    # return the verification key derived from the get_signing_key signing key
    return get_signing_key().verify_key


def test_create_signed_message() -> None:
    """Tests that SignedMessage can be created and serialized"""

    # we will be signing this serializable object
    msg = get_repr_message()
    signing_key = get_signing_key()
    sig_msg = msg.sign(signing_key=signing_key)

    assert sig_msg.obj_type == get_fully_qualified_name(obj=msg)
    assert (
        str(sig_msg.obj_type)
        == "syft.core.node.common.service.repr_service.ReprMessage"
    )
    assert type(sig_msg.signature) == bytes

    assert get_fully_qualified_name(obj=msg) in str(type(sig_msg.message))
    assert len(sig_msg.signature) > 0
    assert sig_msg.verify_key == get_verify_key()
    assert sig_msg.message == msg
    assert sig_msg.serialized_message == serialize(msg, to_bytes=True)


def test_deserialize_signed_message() -> None:
    """Tests that SignedMessage can be deserialized"""

    sig_msg_blob = get_signed_message_bytes()
    sig_msg = sy.deserialize(blob=sig_msg_blob, from_bytes=True)

    msg = get_repr_message()
    signing_key = get_signing_key()
    sig_msg_comp = msg.sign(signing_key=signing_key)

    assert sig_msg.obj_type == sig_msg_comp.obj_type
    assert (
        str(sig_msg.obj_type)
        == "syft.core.node.common.service.repr_service.ReprMessage"
    )

    assert type(sig_msg.signature) == bytes
    assert type(sig_msg.verify_key) == VerifyKey
    assert sig_msg.signature == sig_msg_comp.signature
    assert sig_msg.verify_key == get_verify_key()
    assert sig_msg.message == msg
    assert sig_msg.message == sig_msg_comp.message
    assert sig_msg.serialized_message == sig_msg_comp.serialized_message


def test_serde_matches() -> None:
    """Tests that the nested serde is reversible at all levels"""

    # serial
    sig_msg_blob = get_signed_message_bytes()

    # deserial should be expected type
    sig_msg = sy.deserialize(blob=sig_msg_blob, from_bytes=True)
    assert type(sig_msg) == SignedImmediateSyftMessageWithoutReply

    # reserial should be same as original fixture
    comp_blob = serialize(sig_msg, to_bytes=True)
    assert type(comp_blob) == bytes
    assert comp_blob == sig_msg_blob

    # now try sub message
    msg = sig_msg.message
    assert type(msg) == ReprMessage

    # resign and the result should be the same
    signing_key = get_signing_key()
    sig_msg_comp = msg.sign(signing_key=signing_key)
    assert type(sig_msg_comp) == SignedImmediateSyftMessageWithoutReply
    assert type(sig_msg_comp) == type(sig_msg)

    # make sure they have the same message id for comparison
    sig_msg_comp._id = sig_msg.id
    # identical (except the auto generated UID for the envelope)
    assert sig_msg_comp == sig_msg


def test_verify_message() -> None:
    """Tests that SignedMessage can be verified"""

    blob = get_signed_message_bytes()
    sig_msg = sy.deserialize(blob=blob, from_bytes=True)

    veri_msg = sig_msg.message
    obj = get_repr_message()

    assert veri_msg == obj


def test_verify_message_fails_key() -> None:
    """Tests that SignedMessage cant be verified with the wrong verification key"""

    blob = get_signed_message_bytes()
    sig_msg = sy.deserialize(blob=blob, from_bytes=True)

    # everything is good
    assert sig_msg.is_valid is True

    # change verify_key
    signing_key = SigningKey.generate()
    sig_msg.verify_key = signing_key.verify_key

    # not so good
    assert sig_msg.is_valid is False


def test_verify_message_fails_sig() -> None:
    """Tests that SignedMessage cant be verified with the wrong signature"""

    blob = get_signed_message_bytes()
    sig_msg = sy.deserialize(blob=blob, from_bytes=True)

    # everything is good
    assert sig_msg.is_valid is True

    # change signature
    sig_msg.signature += b"a"

    # not so good
    assert sig_msg.is_valid is False


def test_verify_message_fails_message() -> None:
    """Tests that SignedMessage cant be verified with the wrong message"""

    blob = get_signed_message_bytes()
    sig_msg = sy.deserialize(blob=blob, from_bytes=True)

    # everything is good
    assert sig_msg.is_valid is True

    # change message
    sig_msg.serialized_message += b"a"

    # not so good
    assert sig_msg.is_valid is False


def test_verify_message_fails_empty() -> None:
    """Tests that SignedMessage cant be verified with empty sig"""

    blob = get_signed_message_bytes()
    sig_msg = sy.deserialize(blob=blob, from_bytes=True)

    # everything is good
    assert sig_msg.is_valid is True

    # change message
    sig_msg.signature = b""

    # not so good
    assert sig_msg.is_valid is False


def test_decode_message() -> None:
    """Tests that SignedMessage serialized_message is not encrypted"""

    blob = get_signed_message_bytes()
    sig_msg = sy.deserialize(blob=blob, from_bytes=True)

    nonveri_msg = sy.deserialize(blob=sig_msg.serialized_message, from_bytes=True)
    obj = get_repr_message()

    assert nonveri_msg == obj


def test_get_message() -> None:
    """Tests that SignedMessage verification can be ignored"""

    blob = get_signed_message_bytes()
    sig_msg = sy.deserialize(blob=blob, from_bytes=True)

    sig_msg.signature += b"a"
    nonveri_msg = sig_msg.message
    obj = get_repr_message()

    assert nonveri_msg == obj



# stdlib
from typing import Any
from typing import Optional
from typing import Union

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from nacl.signing import SigningKey
from nacl.signing import VerifyKey

# syft relative
from ...logger import debug
from ...logger import traceback_and_raise
from ...proto.core.io.address_pb2 import Address as Address_PB
from ...util import key_emoji as key_emoji_util
from ..common.serde.deserialize import _deserialize
from ..common.serde.serializable import Serializable
from ..common.serde.serializable import bind_protobuf
from ..common.serde.serialize import _serialize as serialize
from ..common.uid import UID
from ..io.location import Location


class Unspecified(object):
    def __repr__(self) -> str:
        return "Unspecified"


@bind_protobuf
class Address(Serializable):
    name: Optional[str]

    def __init__(
        self,
        name: Optional[str] = None,
        network: Optional[Location] = None,
        domain: Optional[Location] = None,
        device: Optional[Location] = None,
        vm: Optional[Location] = None,
    ):
        self.name = name if name is not None else Serializable.random_name()

        # All node should have a representation of where they think
        # they are currently held. Note that this is at risk of going
        # out of date and so we need to make sure we write good
        # logic to keep these addresses up to date. The main
        # way that it could go out of date is by the node being moved
        # by its parent or its parent being moved by a grandparent, etc.
        # without anyone telling this node. This would be bad because
        # it would mean that when the node creates a new Client for
        # someone to use, it might have trouble actually reaching
        # the node. Fortunately, the creation of a client is (always?)
        # going to be initiated by the parent node itself, so we should
        # be able to check for it there. TODO: did we check for it?

        # this address points to a node, if that node lives within a network,
        # or is a network itself, this property will store the ID of that network
        # if it is known.
        self._network = network

        # this address points to a node, if that node lives within a domain
        # or is a domain itself, this property will store the ID of that domain
        # if it is known.
        self._domain = domain

        # this address points to a node, if that node lives within a device
        # or is a device itself, this property will store the ID of that device
        # if it is known
        self._device = device

        # this client points to a node, if that node lives within a vm
        # or is a vm itself, this property will store the ID of that vm if it
        # is known
        self._vm = vm

    @property
    def icon(self) -> str:
        # 4 different aspects of location
        icon = "💠"
        sub = []
        if self.vm is not None:
            sub.append("🍰")
        if self.device is not None:
            sub.append("📱")
        if self.domain is not None:
            sub.append("🏰")
        if self.network is not None:
            sub.append("🔗")

        if len(sub) > 0:
            icon = f"{icon} ["
            for s in sub:
                icon += s
            icon += "]"
        return icon

    @property
    def pprint(self) -> str:
        output = f"{self.icon} {self.named} ({self.class_name})"
        if hasattr(self, "id"):
            output += f"@{self.target_id.id.emoji()}"
        return output

    def post_init(self) -> None:
        debug(f"> Creating {self.pprint}")

    def key_emoji(self, key: Union[bytes, SigningKey, VerifyKey]) -> str:
        return key_emoji_util(key=key)

    @property
    def address(self) -> "Address":
        # QUESTION what happens if we have none of these?

        # sneak the name on there
        if hasattr(self, "name"):
            name = self.name
        else:
            name = Serializable.random_name()

        address = Address(
            name=name,
            network=self.network,
            domain=self.domain,
            device=self.device,
            vm=self.vm,
        )

        return address

    def _object2proto(self) -> Address_PB:
        """Returns a protobuf serialization of self.

        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.

        :return: returns a protobuf object
        :rtype: Address_PB

        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return Address_PB(
            name=self.name,
            has_network=self.network is not None,
            network=serialize(self.network) if self.network is not None else None,
            has_domain=self.domain is not None,
            domain=serialize(self.domain) if self.domain is not None else None,
            has_device=self.device is not None,
            device=serialize(self.device) if self.device is not None else None,
            has_vm=self.vm is not None,
            vm=serialize(self.vm) if self.vm is not None else None,
        )

    @staticmethod
    def _proto2object(proto: Address_PB) -> "Address":
        """Creates a ObjectWithID from a protobuf

        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.

        :return: returns an instance of ObjectWithID
        :rtype: ObjectWithID

        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """
        return Address(
            name=proto.name,
            network=_deserialize(blob=proto.network) if proto.has_network else None,
            domain=_deserialize(blob=proto.domain) if proto.has_domain else None,
            device=_deserialize(blob=proto.device) if proto.has_device else None,
            vm=_deserialize(blob=proto.vm) if proto.has_vm else None,
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for details.

        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType

        """

        return Address_PB

    @property
    def network(self) -> Optional[Location]:
        """This client points to a node, if that node lives within a network
        or is a network itself, this property will return the ID of that network
        if it is known by the client."""

        return self._network

    @network.setter
    def network(self, new_network: Location) -> Optional[Location]:
        """This client points to a node, if that node lives within a network
        or is a network itself and we learn the id of that network, this setter
        allows us to save the id of that network for use later. We use a getter
        (@property) and setter (@set) explicitly because we want all clients
        to efficiently save an address object for use when sending messages. That
        address object will include this information if it is available"""
        self._network = new_network
        return self._network

    @property
    def network_id(self) -> Optional[UID]:
        network = self.network
        if network is not None:
            return network.id
        return None

    @property
    def domain(self) -> Optional[Location]:
        """This client points to a node, if that node lives within a domain
        or is a domain itself, this property will return the ID of that domain
        if it is known by the client."""

        return self._domain

    @domain.setter
    def domain(self, new_domain: Location) -> Optional[Location]:
        """This client points to a node, if that node lives within a domain
        or is a domain itself and we learn the id of that domain, this setter
        allows us to save the id of that domain for use later. We use a getter
        (@property) and setter (@set) explicitly because we want all clients
        to efficiently save an address object for use when sending messages to their
        target. That address object will include this information if it is available"""

        self._domain = new_domain
        return self._domain

    @property
    def domain_id(self) -> Optional[UID]:
        domain = self.domain
        if domain is not None:
            return domain.id
        return None

    @property
    def device(self) -> Optional[Location]:
        """This client points to a node, if that node lives within a device
        or is a device itself, this property will return the ID of that device
        if it is known by the client."""
        return self._device

    @device.setter
    def device(self, new_device: Location) -> Optional[Location]:
        """This client points to a node, if that node lives within a device
        or is a device itself and we learn the id of that device, this setter
        allows us to save the id of that device for use later. We use a getter
        (@property) and setter (@set) explicitly because we want all clients
        to efficiently save an address object for use when sending messages to their
        target. That address object will include this information if it is available"""

        self._device = new_device
        return self._device

    @property
    def device_id(self) -> Optional[UID]:
        device = self.device
        if device is not None:
            return device.id
        return None

    @property
    def vm(self) -> Optional[Location]:
        """This client points to a node, if that node lives within a vm
        or is a vm itself, this property will return the ID of that vm
        if it is known by the client."""

        return self._vm

    @vm.setter
    def vm(self, new_vm: Location) -> Optional[Location]:
        """This client points to a node, if that node lives within a vm
        or is a vm itself and we learn the id of that vm, this setter
        allows us to save the id of that vm for use later. We use a getter
        (@property) and setter (@set) explicitly because we want all clients
        to efficiently save an address object for use when sending messages to their
        target. That address object will include this information if it is available"""

        self._vm = new_vm
        return self._vm

    @property
    def vm_id(self) -> Optional[UID]:
        vm = self.vm
        if vm is not None:
            return vm.id
        return None

    def target_emoji(self) -> str:
        output = ""
        if self.target_id is not None:
            output = f"@{self.target_id.id.emoji()}"
        return output

    @property
    def target_id(self) -> Location:
        """Return the address of the node which lives at this address.

        Note that this id is simply the most granular id available to the
        address."""
        if self._vm is not None:
            return self._vm
        elif self._device is not None:
            return self._device
        elif self._domain is not None:
            return self._domain
        elif self._network is not None:
            return self._network

        traceback_and_raise(Exception("Address has no valid parts"))

    def __eq__(self, other: Any) -> bool:
        """Returns whether two Address objects refer to the same set of locations

        :param other: the other object to compare with self
        :type other: Any (note this must be Any or __eq__ fails on other types)
        :returns: whether the two objects are the same
        :rtype: bool
        """

        try:
            a = self.network == other.network
            b = self.domain == other.domain
            c = self.device == other.device
            d = self.vm == other.vm

            return a and b and c and d
        except Exception:
            return False

    def __repr__(self) -> str:
        out = f"<{type(self).__name__} -"
        if self.network is not None:
            out += f" Network:{self.network.repr_short()},"  # OpenGrid
        if self.domain is not None:
            out += f" Domain:{self.domain.repr_short()} "  # UCSF
        if self.device is not None:
            out += f" Device:{self.device.repr_short()},"  # One of UCSF's Dell Servers
        if self.vm is not None:
            out += f" VM:{self.vm.repr_short()}"  # 8GB RAM set aside @Trask - UCSF-Server-5

        # remove extraneous comma and add a close carrot
        return out[:-1] + ">"



# third party
import torch as th

# syft absolute
import syft as sy
from syft.lib.python.slice import Slice
from syft.proto.lib.python.slice_pb2 import Slice as Slice_PB


def test_slice_serde() -> None:
    syft_slice = Slice(1, 3, -1)
    serialized = syft_slice._object2proto()

    assert isinstance(serialized, Slice_PB)

    deserialized = Slice._proto2object(proto=serialized)

    assert isinstance(deserialized, Slice)
    assert deserialized.id == syft_slice.id
    assert deserialized.start == syft_slice.start
    assert deserialized.stop == syft_slice.stop
    assert deserialized.step == syft_slice.step


def test_slice_send(client: sy.VirtualMachineClient) -> None:
    syft_slice = Slice(1, 3, None)
    ptr = syft_slice.send(client)

    # Check pointer type
    assert ptr.__class__.__name__ == "SlicePointer"

    # Check that we can get back the object
    res = ptr.get()
    assert res.start == syft_slice.start
    assert res.stop == syft_slice.stop
    assert res.step == syft_slice.step


def test_slice_tensor(client) -> None:
    syft_slice = Slice(0, 1)
    slice_ptr = syft_slice.send(client)

    t = th.Tensor([1, 2, 3])
    t_ptr = t.send(client)
    res_ptr = t_ptr[slice_ptr]

    # Check that we can get back the object
    res = res_ptr.get()
    assert res == t[0:1]

    res_ptr2 = t_ptr[0:1]
    res2 = res_ptr2.get()

    assert res == res2

    last_ptr = t_ptr[-1]
    last = last_ptr.item().get()
    assert last == 3



# -*- coding: utf-8 -*-
"""
This is a skeleton file that can serve as a starting point for a Python
console script. To run this script uncomment the following lines in the
[options.entry_points] section in setup.cfg:
    console_scripts =
         fibonacci = syft.skeleton:run
Then run `python setup.py install` which will install the command `fibonacci`
inside your current environment.
Besides console scripts, the header (i.e. until _logger...) of this file can
also be used as template for Python modules.
Note: This skeleton file can be safely removed if not needed!
"""
# stdlib
from typing import List as TypeList

# third party
import torch

module_type = type(torch)
func_type = type(lambda x: x)
builtin_func_type = type(torch.ones)
class_type = type(func_type)


def unsplit(list_of_things: TypeList[str], separator: str = ".") -> str:
    return separator.join(list_of_things)



"""In this test suite, we evaluate the MemoryStore class. For more info
on the MemoryStore class and its purpose, please see the documentation
in the class itself.

Table of Contents:
    - INITIALIZATION: tests for ways MemoryStore can be initialized
    - CLASS METHODS: tests for the use of MemoryStore's class methods
"""

# stdlib
import re
from typing import List
from typing import Tuple

# third party
import torch as th

# syft absolute
from syft.core.common import UID
from syft.core.common.object import ObjectWithID
from syft.core.store import ObjectStore
from syft.core.store.store_memory import MemoryStore
from syft.core.store.storeable_object import StorableObject


def generate_id_obj(
    data: th.Tensor, description: str, tags: List[str]
) -> Tuple[UID, StorableObject]:
    id = UID()
    obj = StorableObject(id=id, data=data, description=description, tags=tags)

    return id, obj


# --------------------- INITIALIZATION ---------------------


def test_create_memory_storage() -> None:
    """Test that creating MemoryStore() does in fact create
    an ObjectStore."""

    store = MemoryStore()
    assert isinstance(store, ObjectStore)


# --------------------- CLASS METHODS ---------------------


def test_set_item() -> None:
    """Tests that __setitem__ and __getitem__ work intuitively."""

    store = MemoryStore()
    id1, obj1 = generate_id_obj(
        data=th.Tensor([1, 2, 3, 4]),
        description="Dummy tensor",
        tags=["dummy", "tensor"],
    )

    store[id1] = obj1
    assert id1 in store
    assert len(store) == 1


def test_store_item() -> None:
    """Tests that store() works as an alternative to __setitem__."""

    store = MemoryStore()
    id1, obj1 = generate_id_obj(
        data=th.Tensor([1, 2, 3, 4]),
        description="Dummy tensor",
        tags=["dummy", "tensor"],
    )

    store[id1] = obj1
    assert id1 in store
    assert len(store) == 1


def test_get_objects_of_type() -> None:
    """Tests that get_objects_of_type() allows MemoryStore to filter data
    types."""

    store = MemoryStore()
    id1, obj1 = generate_id_obj(
        data=th.Tensor([1, 2, 3, 4]),
        description="Dummy tensor",
        tags=["dummy", "tensor"],
    )
    id2, obj2 = generate_id_obj(
        data=th.Tensor([1, 2, 3]),
        description="Another dummy tensor",
        tags=["another", "dummy", "tensor"],
    )
    id3, obj3 = generate_id_obj(
        data=ObjectWithID(),
        description="Dummy object with ID",
        tags=["dummy", "object", "with", "id"],
    )

    store[id1] = obj1
    store[id2] = obj2
    store[id3] = obj3
    assert all(x in store.get_objects_of_type(th.Tensor) for x in [obj1, obj2])
    assert obj3 in store.get_objects_of_type(ObjectWithID)


def test_keys_values() -> None:
    """Tests that keys() and values() work intuitively and offer MemoryStore
    a dict-like usage."""

    store = MemoryStore()
    id1, obj1 = generate_id_obj(
        data=th.Tensor([1, 2, 3, 4]),
        description="Dummy tensor",
        tags=["dummy", "tensor"],
    )
    id2, obj2 = generate_id_obj(
        data=th.Tensor([1, 2, 3]),
        description="Another dummy tensor",
        tags=["another", "dummy", "tensor"],
    )

    store[id1] = obj1
    store[id2] = obj2
    assert list(store.keys()) == [id1, id2]
    assert list(store.values()) == [obj1, obj2]


def test_clear_len() -> None:
    """Tests that clear() empties the MemoryStore and len() returns the
    number of stored objects."""

    store = MemoryStore()
    id1, obj1 = generate_id_obj(
        data=th.Tensor([1, 2, 3, 4]),
        description="Dummy tensor",
        tags=["dummy", "tensor"],
    )
    id2, obj2 = generate_id_obj(
        data=th.Tensor([1, 2, 3]),
        description="Another dummy tensor",
        tags=["another", "dummy", "tensor"],
    )

    store[id1] = obj1
    store[id2] = obj2

    assert len(store) == 2
    store.clear()
    assert len(store) == 0


def test_str() -> None:
    """Tests that MemoryStore is converted to string properly."""

    store = MemoryStore()
    id1, obj1 = generate_id_obj(
        data=th.Tensor([1, 2, 3, 4]),
        description="Dummy tensor",
        tags=["dummy", "tensor"],
    )

    store[id1] = obj1
    store_regex = re.compile(
        r"^OrderedDict\(\[\(<UID: [0-9a-z]{8}[0-9a-z]{4}[0-9a-z]{4}[0-9a-z]{4}[0-9a-z]"
        + r"{12}>, <Storable: tensor\(\[1\., 2\., 3\., 4\.\]\)>\)\]\)$"
    )
    assert store_regex.match(str(store))



# stdlib
import json
from pathlib import Path
import tempfile
import time
from typing import Any as TypeAny
from typing import Optional as TypeOptional
from typing import Tuple as TypeTuple

# syft relative
from ...logger import debug
from ...logger import info
from ...logger import traceback_and_raise
from .bcolors import bcolors


class DuetCredentialExchanger:
    def __init__(self, *args: TypeTuple[TypeAny, ...], **kwargs: TypeAny) -> None:
        self.join = False

    def set_role(self, join: bool = False) -> "DuetCredentialExchanger":
        self.join = join
        return self

    def set_responder_id(self, credential: TypeAny) -> "DuetCredentialExchanger":
        self.responder_id = credential
        return self

    def run(self, credential: str) -> str:
        traceback_and_raise(NotImplementedError)


class OpenGridTokenManualInputExchanger(DuetCredentialExchanger):
    def run(self, credential: str) -> str:
        self.credential = credential
        if self.join:
            self._client_exchange(credential=self.credential)
            return self.responder_id
        else:
            return self._server_exchange(credential=self.credential)

    def _server_exchange(self, credential: str) -> str:
        # send Server ID
        info(
            "♫♫♫ > Duet Server ID: " + bcolors.BOLD + credential + bcolors.ENDC,
            print=True,
        )

        info(print=True)
        info(
            "♫♫♫ > "
            + bcolors.HEADER
            + "STEP 1:"
            + bcolors.ENDC
            + " Send the following code to your Duet Partner!",
            print=True,
        )
        info("\nimport syft as sy", print=True)
        info(
            'duet = sy.duet("' + bcolors.BOLD + credential + bcolors.ENDC + '")',
            print=True,
        )

        # get Client ID
        info(
            "\n♫♫♫ > "
            + bcolors.HEADER
            + "STEP 2:"
            + bcolors.ENDC
            + " Ask your partner for their Client ID and enter it below!",
            print=True,
        )
        while True:
            client_id = input("♫♫♫ > Duet Partner's Client ID: ")  # nosec
            if len(client_id) == 32:
                break
            else:
                info("    > Error: Invalid Client ID. Please try again.", print=True)
        info(print=True)
        return client_id

    def _client_exchange(self, credential: str) -> None:
        # send client ID
        info(print=True)
        info(
            "♫♫♫ > "
            + bcolors.HEADER
            + "STEP 1:"
            + bcolors.ENDC
            + " Send the following Duet Client ID to your duet partner!",
            print=True,
        )
        info(
            "♫♫♫ > Duet Client ID: " + bcolors.BOLD + credential + bcolors.ENDC,
            print=True,
        )
        info(print=True)
        info("♫♫♫ > ...waiting for partner to connect...", print=True)


def get_loopback_path() -> str:
    loopback_file = "duet_loopback.json"
    return str(Path(tempfile.gettempdir()) / loopback_file)


class OpenGridTokenFileExchanger(DuetCredentialExchanger):
    retry_seconds = 180
    file_path = get_loopback_path()

    def __init__(
        self,
        *args: TypeTuple[TypeAny, ...],
        file_path: TypeOptional[str] = None,
        **kwargs: TypeAny
    ) -> None:
        super().__init__()
        if file_path is not None:
            self.file_path = file_path

    def run(self, credential: str) -> str:
        self.credential = credential
        if self.join:
            return self._client_exchange(credential=self.credential)
        else:
            return self._server_exchange(credential=self.credential)

    def _server_exchange(self, credential: str) -> str:
        info(print=True)
        info(
            "♫♫♫ > "
            + bcolors.HEADER
            + "STEP 1:"
            + bcolors.ENDC
            + " Send the following code to your Duet Partner!",
            print=True,
        )
        info("\nimport syft as sy", print=True)
        info("duet = sy.join_duet(loopback=True)", print=True)
        info(print=True)

        # send Server ID
        loopback_config = {}
        loopback_config["server_id"] = credential
        with open(self.file_path, "w") as f:
            f.write(json.dumps(loopback_config))

        # get Client ID
        client_id = ""
        for retry in range(OpenGridTokenFileExchanger.retry_seconds):
            try:
                with open(self.file_path, "r") as f:
                    loopback_config = json.loads(f.read())

                    if "client_id" in loopback_config:
                        client_id = str(loopback_config["client_id"])
                        break
                debug("client not ready")
                time.sleep(0.5)
            except KeyboardInterrupt:
                debug("Cancelling server connection")
                break
            except Exception as e:
                info("server config load failed", self.file_path, e)
                time.sleep(0.5)

        if client_id == "":
            raise Exception("failed to load client ID")

        return client_id

    def _client_exchange(self, credential: str) -> str:
        loopback_config = {}
        server_id = ""
        for retry in range(OpenGridTokenFileExchanger.retry_seconds):
            try:
                with open(self.file_path, "r") as f:
                    loopback_config = json.loads(f.read())
                    # only continue once the server has overwritten the file
                    # with only its new server_id
                    if (
                        "server_id" in loopback_config
                        and "client_id" not in loopback_config
                    ):
                        server_id = str(loopback_config["server_id"])
                        break
                debug("server not ready")
                time.sleep(0.5)
            except KeyboardInterrupt:
                debug("Cancelling client connection")
                break
            except Exception as e:
                info("client config load failed", self.file_path, e)
                time.sleep(0.5)

        if server_id == "":
            raise Exception("failed to load client ID")

        loopback_config["client_id"] = credential

        with open(self.file_path, "w") as f:
            f.write(json.dumps(loopback_config))

        return server_id



# stdlib
import inspect
from typing import Any
from typing import Callable
from typing import Dict
from typing import Tuple

# syft relative
from ...core.node.common.client import Client
from ...core.node.vm.plan_vm import PlanVirtualMachine
from ...core.pointer.pointer import Pointer
from ...logger import info
from ...logger import traceback_and_raise
from .plan import Plan

PLAN_BUILDER_VM: PlanVirtualMachine = PlanVirtualMachine(name="plan_vm")
ROOT_CLIENT: Client = PLAN_BUILDER_VM.get_root_client()


def build_plan_inputs(forward_func: Callable) -> Dict[str, Pointer]:
    signature = inspect.signature(forward_func)
    res = {}
    for k, v in signature.parameters.items():
        if v.default is not inspect.Parameter.empty:
            res[k] = v.default.send(ROOT_CLIENT)
        else:
            raise ValueError(
                f"arg {k} has no placeholder as default value (required for @make_plan functions)"
            )
    return res


def make_plan(func: Callable) -> Plan:
    inputs = build_plan_inputs(func)
    vm = PLAN_BUILDER_VM
    code = inspect.getsource(func)
    vm.record_actions()
    res = func(**inputs)
    vm.stop_recording()
    plan = Plan(actions=vm.recorded_actions, inputs=inputs, outputs=res, code=code)
    # cleanup
    vm.recorded_actions = []
    return plan


class PlanBuilder:
    def __init__(self, vm: PlanVirtualMachine = PLAN_BUILDER_VM):
        self.vm = vm

        try:
            self.plan: Plan = self.build()
        except Exception as e:
            traceback_and_raise(ValueError(f"Failed to build plan. {e}."))

        self.plan_pointer = None

    def build(self, *args: Tuple[Any, ...]) -> Plan:
        inputs = build_plan_inputs(self.forward)

        self.vm.record_actions()
        res = self.forward(**inputs)
        self.vm.stop_recording()
        plan: Plan = Plan(actions=self.vm.recorded_actions, inputs=inputs, outputs=res)
        self.vm.recorded_actions = []
        return plan

    def forward(self, *args: Tuple[Any, ...], **kwargs: Any) -> Any:
        raise NotImplementedError

    def __call__(self, **kwargs: Any) -> Any:
        if self.plan_pointer is None:
            client: Client = PlanVirtualMachine(name="alice").get_root_client()
            info("Model is not remote yet, sending to a new VM")
            self.send_plan(client)

        if self.plan_pointer is None:
            traceback_and_raise(ValueError("Unable to call Plan. Send failed."))

        return self.plan_pointer(**kwargs)  # type: ignore

    def send_plan(self, client: Client) -> None:
        self.plan_pointer = self.plan.send(client)  # type: ignore



# stdlib
from collections import OrderedDict
from typing import Any

# third party
import pytest

# syft absolute
import syft as sy


def get_permission(
    obj: Any, node: sy.VirtualMachine, client: sy.VirtualMachineClient
) -> None:
    remote_obj = node.store[obj.id_at_location]
    remote_obj.read_permissions[client.verify_key] = obj.id_at_location


inputs = {
    "__contains__": [["1"], ["2"], [1], [2]],
    "__delitem__": [["1"], [1], [10], ["10"], [500], [{1: 1}]],
    "__eq__": [
        [OrderedDict([("1", 1), ("2", 2), ("3", 3)])],
        [OrderedDict([("1", 1), ("2", 2), ("3", 3), ("4", 4)])],
    ],
    "__getitem__": [["1"], [1], [10], ["10"], [500], [{1: 1}]],
    "__len__": [[]],
    "__ne__": [
        [OrderedDict([("1", 1), ("2", 2), ("3", 3)])],
        [OrderedDict([("1", 1), ("2", 2), ("3", 3), ("4", 4)])],
    ],
    "__setitem__": [["1", None], [1, None], [10, None], ["10", 10], [500, None]],
    "clear": [[]],
    "copy": [[]],
    "fromkeys": [[[1, 2, 3]], [["1", 2, "4"]]],
    "get": [[1], [2], [4], [-1]],
    "items": [[]],
    "keys": [[]],
    "move_to_end": [[1], [2], [50], [-1]],
    "pop": [[1], [2], [50], [-1]],
    "popitem": [[True], [False]],
    "setdefault": [[1, "1"], [-1, 1], ["1", 1]],
    "update": [[{1: 1}], [{1: 1, 2: 2, -1: 5}]],
    "values": [[]],
}

objects = [
    [("1", 1), ("2", 2), ("3", 3)],
    list(zip(range(100), range(100))),
]


@pytest.mark.slow
@pytest.mark.parametrize("test_object", objects)
@pytest.mark.parametrize("func", inputs.keys())
def test_pointer_objectives(
    test_object, func, node: sy.VirtualMachine, client: sy.VirtualMachineClient
):
    py_obj, sy_obj, remote_sy_obj = (
        OrderedDict(test_object),
        sy.lib.python.collections.OrderedDict(test_object),
        client.syft.lib.python.collections.OrderedDict(test_object),
    )

    possible_inputs = inputs[func]

    if not hasattr(py_obj, func):
        return

    py_method = getattr(py_obj, func)

    if func == "get":
        func = "dict_get"

    sy_method = getattr(sy_obj, func)

    if func == "__len__":
        func = "len"

    remote_sy_method = getattr(remote_sy_obj, func)

    for possible_input in possible_inputs:
        try:
            py_res = py_method(*possible_input)
        except Exception as py_e:
            py_res = str(py_e)

        try:
            sy_res = sy_method(*possible_input)
        except Exception as sy_e:
            sy_res = str(sy_e)

        try:
            remote_sy_res = remote_sy_method(*possible_input)
            get_permission(remote_sy_res, node, client)
            remote_sy_res = remote_sy_res.get()
        except Exception as remote_sy_e:
            remote_sy_res = str(remote_sy_e)

        if isinstance(py_res, float):
            py_res = int(py_res * 1000) / 1000
            sy_res = int(sy_res * 1000) / 1000
            remote_sy_res = int(remote_sy_res * 1000) / 1000

        if func in ["items", "values", "keys"]:
            py_res = list(py_res)
            sy_res = list(sy_res)

        assert py_res == sy_res

        # TODO: support `.get` for IteratorPointer objects
        if func not in ("items", "keys", "values"):
            assert sy_res == remote_sy_res



"""
WebRTC connection representation

This class aims to represent a generic and
asynchronous peer-to-peer WebRTC connection
based in Syft BidirectionalConnection Interface.

This connection interface provides a full-duplex
channel, allowing this class to work as a client
and as a server at the same time.

This class is useful to send/compute data
using a p2p channel without the need
of deploying servers on cloud providers
or setting firewalls rules  in order
turn this process "visible" to the world.

How does it work?
The WebRTC (Web Real-Time Communication)
is a protocol that uses a full-duplex channel
(same API as WebSockets), in order to provide
a high-quality RTC, and NAT traversal
networking technique in order to be able
to reach private addresses.

The WebRTC API includes no provisions
for signaling, the applications are in charge
of managing how the connections will be established.
In our use case, we'll be using ICE (Interactive
Connectivity Establishment) to establish these connections.

Signaling Components:
    - PyGrid Network App (Signaling Server)
    - PySyft Peer Process (Offer)
    - PySyft Peer Process (Answer)

Signaling Steps:

    1 - [PUSH] PySyft Peer (Offer) will send an offer msg
    to the Signaling Server containing its own local
    description (IP, MAC address, etc), addressing (PySyft.Address)
    it to the target node.

    2 - The Signaling Server (PyGrid Network) will
    receive the offer msg and store it in the target's
    offer request's queue.

    3 - [PULL] The PySyft Peer (Answer) will send a message to
    the Signaling Server checking if the desired node pushed
    any offer msg in their queue.

    4 - The Signaling Server will check the existence of offer messages addressed
    to the PySyft Peer (Answer) made by the desired node address (PySyft.Address).
    If that's the case, then the offer message will be sent to the peer as a response.

    5 - [PUSH] The PySyft Peer (Answer) will process the offer message
    in order to know the network address of the other peer.
    If no problems were found, the answering peer will generate its own local
    description (IP, Mac address,  etc) and send it to the signaling server,
    addressing (PySyft.Address) it to the offering node as an AnswerMessage.

    6 - The Signaling Server (PyGrid Network) will
    receive the answer msg and store it in the target's answer request's
    queue.

    7 - [PULL] The PySyft Peer (Offer) will send a message to
    the Signaling Server checking if the desired node pushed
    any answer msg in his queue.

    8 - The Signaling Server will check the existence of answer messages addressed
    to the PySyft Peer (Offer) made by the desired node address (PySyft.Address).
    If that's the case, so the answer message will be sent to the peer as a response.

    9 - The PySyft Peer (Offer) will process the answer message
    in order to know the network address of the other peer.
    If no problems were found, the connection channel will be established.
"""

# stdlib
import asyncio
import math
import os
import secrets
import time
from typing import Any
from typing import Optional
from typing import Union

# third party
from aiortc import RTCDataChannel
from aiortc import RTCPeerConnection
from aiortc import RTCSessionDescription
from aiortc.contrib.signaling import object_from_string
from aiortc.contrib.signaling import object_to_string

# syft relative
from ... import serialize
from ...core.common.event_loop import loop
from ...core.common.message import SignedEventualSyftMessageWithoutReply
from ...core.common.message import SignedImmediateSyftMessageWithReply
from ...core.common.message import SignedImmediateSyftMessageWithoutReply
from ...core.common.serde.deserialize import _deserialize
from ...core.io.address import Address
from ...core.io.connection import BidirectionalConnection
from ...core.node.abstract.node import AbstractNode
from ...logger import debug
from ...logger import traceback_and_raise
from ...util import validate_type
from ..services.signaling_service import CloseConnectionMessage

DC_CHUNKING_ENABLED = True
DC_CHUNK_START_SIGN = b"<<<CHUNK START>>>"

try:
    DC_MAX_CHUNK_SIZE = int(os.environ["DC_MAX_CHUNK_SIZE"])
except KeyError:
    DC_MAX_CHUNK_SIZE = 2 ** 18

try:
    DC_MAX_BUFSIZE = int(os.environ["DC_MAX_BUFSIZE"])
except KeyError:
    DC_MAX_BUFSIZE = 2 ** 22


class OrderedChunk:
    def __init__(self, idx: int, data: bytes):
        self.idx = idx
        self.data = data

    def save(self) -> bytes:
        return self.idx.to_bytes(4, "big") + self.data

    @classmethod
    def load(cls, data: bytes) -> "OrderedChunk":
        idx = int.from_bytes(data[:4], "big")
        data = data[4:]
        return cls(idx, data)


class WebRTCConnection(BidirectionalConnection):
    loop: Any

    def __init__(self, node: AbstractNode) -> None:
        # WebRTC Connection representation

        # As we have a full-duplex connection,
        # it's necessary to use a node instance
        # inside of this connection. In order to
        # be able to process requests sent by
        # the other peer.
        # All the requests messages will be forwarded
        # to this node.
        self.node = node

        # EventLoop that manages async tasks (producer/consumer)
        # This structure is global and needs to be
        # defined beforehand.

        self.loop = loop
        # Message pool (High Priority)
        # These queues will be used to manage
        # async  messages.
        try:
            self.producer_pool: asyncio.Queue = asyncio.Queue(
                loop=self.loop,
            )  # Request Messages / Request Responses
            self.consumer_pool: asyncio.Queue = asyncio.Queue(
                loop=self.loop,
            )  # Request Responses

            # Initialize a PeerConnection structure
            self.peer_connection = RTCPeerConnection()

            # Set channel descriptor as None
            # This attribute will be used for external classes
            # in order to verify if the connection channel
            # was established.
            self.channel: RTCDataChannel
            self._client_address: Optional[Address] = None

        except Exception as e:
            traceback_and_raise(e)

    async def _set_offer(self) -> str:
        """
        Initialize a Real-Time Communication Data Channel,
        set data channel callbacks/tasks, and send offer payload
        message.

        :return: returns a signaling offer payload containing local description.
        :rtype: str
        """
        try:
            # Use the Peer Connection structure to
            # set the channel as a RTCDataChannel.
            self.channel = self.peer_connection.createDataChannel(
                "datachannel",
            )
            # Keep send buffer busy with chunks
            self.channel.bufferedAmountLowThreshold = 4 * DC_MAX_CHUNK_SIZE

            # This method will be called by aioRTC lib as a callback
            # function when the connection opens.
            @self.channel.on("open")
            async def on_open() -> None:  # type : ignore
                self.__producer_task = asyncio.ensure_future(self.producer())

            chunked_msg = []
            chunks_pending = 0

            # This method is the aioRTC "consumer" task
            # and will be running as long as connection remains.
            # At this point we're just setting the method behavior
            # It'll start running after the connection opens.
            @self.channel.on("message")
            async def on_message(raw: bytes) -> None:
                nonlocal chunked_msg, chunks_pending

                chunk = OrderedChunk.load(raw)
                message = chunk.data

                if message == DC_CHUNK_START_SIGN:
                    chunks_pending = chunk.idx
                    chunked_msg = [b""] * chunks_pending
                elif chunks_pending:
                    if chunked_msg[chunk.idx] == b"":
                        chunks_pending -= 1
                    chunked_msg[chunk.idx] = message
                    if chunks_pending == 0:
                        await self.consumer(msg=b"".join(chunked_msg))
                else:
                    # Forward all received messages to our own consumer method.
                    await self.consumer(msg=message)

            # Set peer_connection to generate an offer message type.
            await self.peer_connection.setLocalDescription(
                await self.peer_connection.createOffer()
            )

            # Generates the local description structure
            # and serialize it to string afterwards.
            local_description = object_to_string(self.peer_connection.localDescription)

            # Return the Offer local_description payload.
            return local_description
        except Exception as e:
            traceback_and_raise(e)

    async def _set_answer(self, payload: str) -> str:
        """
        Receives a signaling offer payload, initialize/set
        Data channel callbacks/tasks, updates remote local description
        using offer's payload message and returns a
        signaling answer payload.

        :return: returns a signaling answer payload containing local description.
        :rtype: str
        """

        try:

            @self.peer_connection.on("datachannel")
            def on_datachannel(channel: RTCDataChannel) -> None:
                self.channel = channel

                self.__producer_task = asyncio.ensure_future(self.producer())

                chunked_msg = []
                chunks_pending = 0

                @self.channel.on("message")
                async def on_message(raw: bytes) -> None:
                    nonlocal chunked_msg, chunks_pending

                    chunk = OrderedChunk.load(raw)
                    message = chunk.data
                    if message == DC_CHUNK_START_SIGN:
                        chunks_pending = chunk.idx
                        chunked_msg = [b""] * chunks_pending
                    elif chunks_pending:
                        if chunked_msg[chunk.idx] == b"":
                            chunks_pending -= 1
                        chunked_msg[chunk.idx] = message
                        if chunks_pending == 0:
                            await self.consumer(msg=b"".join(chunked_msg))
                    else:
                        await self.consumer(msg=message)

            result = await self._process_answer(payload=payload)
            return validate_type(result, str)

        except Exception as e:
            traceback_and_raise(e)
            raise Exception("mypy workaound: should not get here")

    async def _process_answer(self, payload: str) -> Union[str, None]:
        # Converts payload received by
        # the other peer in aioRTC Object
        # instance.
        try:
            msg = object_from_string(payload)

            # Check if Object instance is a
            # description of RTC Session.
            if isinstance(msg, RTCSessionDescription):

                # Use the target's network address/metadata
                # to set the remote description of this peer.
                # This will basically say to this peer how to find/connect
                # with to other peer.
                await self.peer_connection.setRemoteDescription(msg)

                # If it's an offer message type,
                # generates your own local description
                # and send it back in order to tell
                # to the other peer how to find you.
                if msg.type == "offer":
                    # Set peer_connection to generate an offer message type.
                    await self.peer_connection.setLocalDescription(
                        await self.peer_connection.createAnswer()
                    )

                    # Generates the local description structure
                    # and serialize it to string afterwards.
                    local_description = object_to_string(
                        self.peer_connection.localDescription
                    )

                    # Returns the answer peer's local description
                    return local_description
        except Exception as e:
            traceback_and_raise(e)
        return None

    async def producer(self) -> None:
        """
        Async task to send messages to the other side.
        These messages will be enqueued by PySyft Node Clients
        by using PySyft routes and ClientConnection's inheritance.
        """
        try:
            while True:
                # If self.producer_pool is empty, give up task queue priority
                # and give computing time to the next task.
                msg = await self.producer_pool.get()

                # If self.producer_pool.get() returns a message
                # send it as a binary using the RTCDataChannel.
                data = serialize(msg, to_bytes=True)
                data_len = len(data)

                if DC_CHUNKING_ENABLED and data_len > DC_MAX_CHUNK_SIZE:
                    chunk_num = 0
                    done = False
                    sent: asyncio.Future = asyncio.Future(loop=self.loop)

                    def send_data_chunks() -> None:
                        nonlocal chunk_num, data_len, done, sent
                        # Send chunks until buffered amount is big or we're done
                        while (
                            self.channel.bufferedAmount <= DC_MAX_BUFSIZE and not done
                        ):
                            start_offset = chunk_num * DC_MAX_CHUNK_SIZE
                            end_offset = min(
                                (chunk_num + 1) * DC_MAX_CHUNK_SIZE, data_len
                            )
                            chunk = data[start_offset:end_offset]
                            self.channel.send(OrderedChunk(chunk_num, chunk).save())
                            chunk_num += 1
                            if chunk_num * DC_MAX_CHUNK_SIZE >= data_len:
                                done = True
                                sent.set_result(True)

                        if not done:
                            # Set listener for next round of sending when buffer is empty
                            self.channel.once("bufferedamountlow", send_data_chunks)

                    chunk_count = math.ceil(data_len / DC_MAX_CHUNK_SIZE)
                    self.channel.send(
                        OrderedChunk(chunk_count, DC_CHUNK_START_SIGN).save()
                    )
                    send_data_chunks()
                    # Wait until all chunks are dispatched
                    await sent
                else:
                    self.channel.send(OrderedChunk(0, data).save())
        except Exception as e:
            traceback_and_raise(e)

    def close(self) -> None:
        try:
            # Build Close Message to warn the other peer
            bye_msg = CloseConnectionMessage(address=Address())

            self.channel.send(OrderedChunk(0, serialize(bye_msg, to_bytes=True)).save())

            # Finish async tasks related with this connection
            self._finish_coroutines()
        except Exception as e:
            traceback_and_raise(e)

    def _finish_coroutines(self) -> None:
        try:
            asyncio.run(self.peer_connection.close())
            self.__producer_task.cancel()
        except Exception as e:
            traceback_and_raise(e)

    async def consumer(self, msg: bytes) -> None:
        """
        Async task to receive/process messages sent by the other side.
        These messages will be sent by the other peer as a service requests or responses
        for requests made by this connection previously (ImmediateSyftMessageWithReply).
        """
        try:
            # Deserialize the received message
            _msg = _deserialize(blob=msg, from_bytes=True)

            # Check if it's NOT  a response generated by a previous request
            # made by the client instance that uses this connection as a route.
            # PS: The "_client_address" attribute will be defined during
            # Node Client initialization.
            if _msg.address != self._client_address:
                # If it's a new service request, route it properly
                # using the node instance owned by this connection.

                # Immediate message with reply
                if isinstance(_msg, SignedImmediateSyftMessageWithReply):
                    reply = self.recv_immediate_msg_with_reply(msg=_msg)
                    await self.producer_pool.put(reply)

                # Immediate message without reply
                elif isinstance(_msg, SignedImmediateSyftMessageWithoutReply):
                    self.recv_immediate_msg_without_reply(msg=_msg)

                elif isinstance(_msg, CloseConnectionMessage):
                    # Just finish async tasks related with this connection
                    self._finish_coroutines()

                # Eventual message without reply
                else:
                    self.recv_eventual_msg_without_reply(msg=_msg)

            # If it's true, the message will have the client's address as destination.
            else:
                await self.consumer_pool.put(_msg)

        except Exception as e:
            traceback_and_raise(e)

    def recv_immediate_msg_with_reply(
        self, msg: SignedImmediateSyftMessageWithReply
    ) -> SignedImmediateSyftMessageWithoutReply:
        """
        Executes/Replies requests instantly.

        :return: returns an instance of SignedImmediateSyftMessageWithReply
        :rtype: SignedImmediateSyftMessageWithoutReply
        """
        # Execute node services now
        try:
            r = secrets.randbelow(100000)
            debug(
                f"> Before recv_immediate_msg_with_reply {r} {msg.message} {type(msg.message)}"
            )
            reply = self.node.recv_immediate_msg_with_reply(msg=msg)
            debug(
                f"> After recv_immediate_msg_with_reply {r} {msg.message} {type(msg.message)}"
            )
            return reply
        except Exception as e:
            traceback_and_raise(e)

    def recv_immediate_msg_without_reply(
        self, msg: SignedImmediateSyftMessageWithoutReply
    ) -> None:
        """
        Executes requests instantly.
        """
        try:
            r = secrets.randbelow(100000)
            debug(
                f"> Before recv_immediate_msg_without_reply {r} {msg.message} {type(msg.message)}"
            )
            self.node.recv_immediate_msg_without_reply(msg=msg)
            debug(
                f"> After recv_immediate_msg_without_reply {r} {msg.message} {type(msg.message)}"
            )
        except Exception as e:
            traceback_and_raise(e)

    def recv_eventual_msg_without_reply(
        self, msg: SignedEventualSyftMessageWithoutReply
    ) -> None:
        """
        Executes requests eventually.
        """
        try:
            self.node.recv_eventual_msg_without_reply(msg=msg)
        except Exception as e:
            traceback_and_raise(e)
            raise Exception("mypy workaound: should not get here")

    # TODO: fix this mypy madness
    def send_immediate_msg_with_reply(  # type: ignore
        self, msg: SignedImmediateSyftMessageWithReply
    ) -> SignedImmediateSyftMessageWithReply:
        """
        Sends high priority messages and wait for their responses.

        :return: returns an instance of SignedImmediateSyftMessageWithReply.
        :rtype: SignedImmediateSyftMessageWithReply
        """
        try:
            # properly fix this!
            return validate_type(
                asyncio.run(self.send_sync_message(msg=msg)),
                object,
            )
        except Exception as e:
            traceback_and_raise(e)
            raise Exception("mypy workaound: should not get here")

    def send_immediate_msg_without_reply(
        self, msg: SignedImmediateSyftMessageWithoutReply
    ) -> None:
        """
        Sends high priority messages without waiting for their reply.
        """
        try:
            # asyncio.run(self.producer_pool.put_nowait(msg))
            self.producer_pool.put_nowait(msg)
        except Exception as e:
            traceback_and_raise(e)

    def send_eventual_msg_without_reply(
        self, msg: SignedEventualSyftMessageWithoutReply
    ) -> None:
        """
        Sends low priority messages without waiting for their reply.
        """
        try:
            asyncio.run(self.producer_pool.put(msg))
        except Exception as e:
            traceback_and_raise(e)

    async def send_sync_message(
        self, msg: SignedImmediateSyftMessageWithReply
    ) -> SignedImmediateSyftMessageWithoutReply:
        """
        Send sync messages generically.

        :return: returns an instance of SignedImmediateSyftMessageWithoutReply.
        :rtype: SignedImmediateSyftMessageWithoutReply
        """
        try:
            # To ensure the sequence of sending / receiving messages
            # it's necessary to keep only a unique reference for reading
            # inputs (producer) and outputs (consumer).
            r = secrets.randbelow(100000)
            # To be able to perform this method synchronously (waiting for the reply)
            # without blocking async methods, we need to use queues.

            # Enqueue the message to be sent to the target.
            debug(f"> Before send_sync_message producer_pool.put blocking {r}")
            # self.producer_pool.put_nowait(msg)
            await self.producer_pool.put(msg)
            debug(f"> After send_sync_message producer_pool.put blocking {r}")

            # Wait for the response checking the consumer queue.
            debug(f"> Before send_sync_message consumer_pool.get blocking {r} {msg}")
            debug(
                f"> Before send_sync_message consumer_pool.get blocking {r} {msg.message}"
            )
            response = await self.consumer_pool.get()

            debug(f"> After send_sync_message consumer_pool.get blocking {r}")
            return response
        except Exception as e:
            traceback_and_raise(e)

    async def async_check(
        self, before: float, timeout_secs: int, r: float
    ) -> SignedImmediateSyftMessageWithoutReply:
        while True:
            try:
                response = self.consumer_pool.get_nowait()
                return response
            except Exception as e:
                now = time.time()
                debug(f"> During send_sync_message consumer_pool.get blocking {r}. {e}")
                if now - before > timeout_secs:
                    traceback_and_raise(
                        Exception(f"send_sync_message timeout {timeout_secs} {r}")
                    )



# run this from the interactive command line like:
# python -i examples/duet/cli/do.py

# or you can run these commands manually in the python REPL

# syft absolute
import syft as sy

duet = sy.launch_duet(loopback=True)
print("DO Store", duet.store)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/lib/torch/module.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.plan import plan_pb2 as proto_dot_core_dot_plan_dot_plan__pb2
from syft.proto.lib.python.collections import (
    ordered_dict_pb2 as proto_dot_lib_dot_python_dot_collections_dot_ordered__dict__pb2,
)

DESCRIPTOR = _descriptor.FileDescriptor(
    name="proto/lib/torch/module.proto",
    package="syft.lib.torch",
    syntax="proto3",
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n\x1cproto/lib/torch/module.proto\x12\x0esyft.lib.torch\x1a\x1aproto/core/plan/plan.proto\x1a/proto/lib/python/collections/ordered_dict.proto"\xaf\x01\n\x06Module\x12\x13\n\x0bmodule_type\x18\x01 \x01(\t\x12\x13\n\x0bmodule_name\x18\x02 \x01(\t\x12\x13\n\x0bmodule_repr\x18\x03 \x01(\t\x12(\n\x08\x63hildren\x18\x04 \x03(\x0b\x32\x16.syft.lib.torch.Module\x12<\n\nstate_dict\x18\x05 \x01(\x0b\x32(.syft.lib.python.collections.OrderedDictb\x06proto3',
    dependencies=[
        proto_dot_core_dot_plan_dot_plan__pb2.DESCRIPTOR,
        proto_dot_lib_dot_python_dot_collections_dot_ordered__dict__pb2.DESCRIPTOR,
    ],
)


_MODULE = _descriptor.Descriptor(
    name="Module",
    full_name="syft.lib.torch.Module",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="module_type",
            full_name="syft.lib.torch.Module.module_type",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="module_name",
            full_name="syft.lib.torch.Module.module_name",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="module_repr",
            full_name="syft.lib.torch.Module.module_repr",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="children",
            full_name="syft.lib.torch.Module.children",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="state_dict",
            full_name="syft.lib.torch.Module.state_dict",
            index=4,
            number=5,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=126,
    serialized_end=301,
)

_MODULE.fields_by_name["children"].message_type = _MODULE
_MODULE.fields_by_name[
    "state_dict"
].message_type = (
    proto_dot_lib_dot_python_dot_collections_dot_ordered__dict__pb2._ORDEREDDICT
)
DESCRIPTOR.message_types_by_name["Module"] = _MODULE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Module = _reflection.GeneratedProtocolMessageType(
    "Module",
    (_message.Message,),
    {
        "DESCRIPTOR": _MODULE,
        "__module__": "proto.lib.torch.module_pb2"
        # @@protoc_insertion_point(class_scope:syft.lib.torch.Module)
    },
)
_sym_db.RegisterMessage(Module)


# @@protoc_insertion_point(module_scope)



# third party
import pytest
import torch as th

# syft absolute
import syft as sy


@pytest.mark.asyncio
@pytest.mark.vendor(lib="sympc")
def test_load_sympc() -> None:
    alice = sy.VirtualMachine()
    alice_client = alice.get_root_client()
    bob = sy.VirtualMachine()
    bob_client = bob.get_root_client()

    # third party
    from sympc.session import Session
    from sympc.session import SessionManager
    from sympc.tensor import MPCTensor

    sy.load("sympc")

    session = Session(parties=[alice_client, bob_client])
    SessionManager.setup_mpc(session)

    y = th.Tensor([-5, 0, 1, 2, 3])
    x_secret = th.Tensor([30])
    x = MPCTensor(secret=x_secret, shape=(1,), session=session)

    assert ((x + y).reconstruct() == th.Tensor([25.0, 30.0, 31.0, 32.0, 33.0])).all()



# stdlib
from typing import Any
from typing import Union

# third party
from google.protobuf.message import Message

# syft relative
from ....logger import traceback_and_raise
from ....proto.util.data_message_pb2 import DataMessage
from ....util import index_syft_by_module_name


def _deserialize(
    blob: Union[str, dict, bytes, Message],
    from_proto: bool = True,
    from_bytes: bool = False,
) -> Any:
    """We assume you're deserializing a protobuf object by default

    This function deserializes from encoding to a Python object. There are a few ways of
    using this function:
    1. An Message object is passed,q this will transform a protobuf message into its associated
    class.
    the from_proto has to be set (it is by default).
    2. Bytes are passed. This requires the from_bytes flag set the schema_type specified.
    We cannot (and we should not) be able to get the schema_type from the binary representation.

    Note: The only format that does not require the schema_type is when we are passing
    Messages directly.

    Raises: ValueError if you are not setting one from_<protocol> flag.
            ValueError if you are deserializing a data type that requires a schema type and not
            providing one.
            TypeError if you are trying to deserialize an unsupported type.

    :param blob: this parameter is the data to be deserialized from various formats.
    :type blob: Union[str, dict, bytes, Messages]
    :param from_proto: set this flag to True if you want to deserialize a protobuf message.
    :param from_bytes: set this flag to True if you want to deserialize a binary object.
    :type from_bytes: bool
    :return: a deserialized form of the object on which _deserialize() is called.
    :rtype: Serializable
    """

    deserialization_error = TypeError(
        "You tried to deserialize an unsupported type. This can be caused by "
        "several reasons. Either you are actively writing Syft code and forgot "
        "to create one, or you are trying to deserialize an object which was "
        "serialized using a different version of Syft and the object you tried "
        "to deserialize is not supported in this version."
    )

    if from_bytes:
        data_message = DataMessage()
        data_message.ParseFromString(blob)
        obj_type = index_syft_by_module_name(fully_qualified_name=data_message.obj_type)
        get_protobuf_schema = getattr(obj_type, "get_protobuf_schema", None)

        if not callable(get_protobuf_schema):
            traceback_and_raise(deserialization_error)

        protobuf_type = get_protobuf_schema()
        blob = protobuf_type()

        if not isinstance(blob, Message):
            traceback_and_raise(deserialization_error)

        blob.ParseFromString(data_message.content)

    # lets try to lookup the type we are deserializing
    obj_type = getattr(type(blob), "schema2type", None)
    # when a protobuf type is related to multiple classes, it's schema2type will be None.
    # In that case, we use it's obj_type field.
    if obj_type is None:
        obj_type = getattr(blob, "obj_type", None)
        if obj_type is None:
            traceback_and_raise(deserialization_error)
        obj_type = index_syft_by_module_name(fully_qualified_name=obj_type)  # type: ignore
        obj_type = getattr(obj_type, "_sy_serializable_wrapper_type", obj_type)

    if not isinstance(obj_type, type):
        traceback_and_raise(deserialization_error)

    _proto2object = getattr(obj_type, "_proto2object", None)
    if not callable(_proto2object):
        traceback_and_raise(deserialization_error)

    res = _proto2object(proto=blob)
    return res



# stdlib
import functools
from typing import Any as TypeAny
from typing import List as TypeList
from typing import Tuple as TypeTuple

# third party
import petlib as pl
import zksk as zk  # noqa: 401 # required for pl.ec and pl.bn

# syft relative
from . import bn  # noqa: 401
from . import ecpt  # noqa: 401
from . import ecpt_group  # noqa: 401
from ...ast import add_classes
from ...ast import add_methods
from ...ast import add_modules
from ...ast.globals import Globals
from ..util import generic_update_ast

LIB_NAME = "petlib"
PACKAGE_SUPPORT = {
    "lib": LIB_NAME,
    "python": {"max_version": (3, 9, 99)},
}


def create_ast(client: TypeAny) -> Globals:
    ast = Globals(client=client)

    modules: TypeList[TypeTuple[str, TypeAny]] = [
        ("petlib", pl),
        ("petlib.ec", pl.ec),
        ("petlib.bn", pl.bn),
    ]

    classes: TypeList[TypeTuple[str, str, TypeAny]] = [
        ("petlib.ec.EcPt", "petlib.ec.EcPt", pl.ec.EcPt),
        ("petlib.ec.EcGroup", "petlib.ec.EcGroup", pl.ec.EcGroup),
        ("petlib.bn.Bn", "petlib.bn.Bn", pl.bn.Bn),
    ]

    methods = [
        ("petlib.ec.EcPt.group", "petlib.ec.EcGroup"),
        ("petlib.ec.EcPt.__copy__", "petlib.ec.EcPt"),
        ("petlib.ec.EcPt.__add__", "petlib.ec.EcPt"),
        ("petlib.ec.EcPt.pt_add", "petlib.ec.EcPt"),
        ("petlib.ec.EcPt.pt_add_inplace", "petlib.ec.EcPt"),
        ("petlib.ec.EcPt.pt_mul", "petlib.ec.EcPt"),
        ("petlib.ec.EcPt.pt_mul_inplace", "petlib.ec.EcPt"),
        ("petlib.ec.EcPt.__rmul__", "petlib.ec.EcPt"),
    ]

    add_modules(ast, modules)
    add_classes(ast, classes)
    add_methods(ast, methods)

    for klass in ast.classes:
        klass.create_pointer_class()
        klass.create_send_method()
        klass.create_storable_object_attr_convenience_methods()

    return ast


update_ast = functools.partial(generic_update_ast, LIB_NAME, create_ast)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/grid/messages/user_messages.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2

DESCRIPTOR = _descriptor.FileDescriptor(
    name="proto/grid/messages/user_messages.proto",
    package="syft.grid.messages",
    syntax="proto3",
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n\'proto/grid/messages/user_messages.proto\x12\x12syft.grid.messages\x1a%proto/core/common/common_object.proto\x1a\x1bproto/core/io/address.proto"\x9c\x01\n\x11\x43reateUserMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x89\x01\n\x12\x43reateUserResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x99\x01\n\x0eGetUserMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x86\x01\n\x0fGetUserResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x9a\x01\n\x0fGetUsersMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x87\x01\n\x10GetUsersResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x9c\x01\n\x11\x44\x65leteUserMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x89\x01\n\x12\x44\x65leteUserResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x9c\x01\n\x11UpdateUserMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x89\x01\n\x12UpdateUserResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x9d\x01\n\x12SearchUsersMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x8a\x01\n\x13SearchUsersResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Addressb\x06proto3',
    dependencies=[
        proto_dot_core_dot_common_dot_common__object__pb2.DESCRIPTOR,
        proto_dot_core_dot_io_dot_address__pb2.DESCRIPTOR,
    ],
)


_CREATEUSERMESSAGE = _descriptor.Descriptor(
    name="CreateUserMessage",
    full_name="syft.grid.messages.CreateUserMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.grid.messages.CreateUserMessage.msg_id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.grid.messages.CreateUserMessage.address",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="content",
            full_name="syft.grid.messages.CreateUserMessage.content",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="reply_to",
            full_name="syft.grid.messages.CreateUserMessage.reply_to",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=132,
    serialized_end=288,
)


_CREATEUSERRESPONSE = _descriptor.Descriptor(
    name="CreateUserResponse",
    full_name="syft.grid.messages.CreateUserResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.grid.messages.CreateUserResponse.msg_id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="status_code",
            full_name="syft.grid.messages.CreateUserResponse.status_code",
            index=1,
            number=2,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="content",
            full_name="syft.grid.messages.CreateUserResponse.content",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.grid.messages.CreateUserResponse.address",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=291,
    serialized_end=428,
)


_GETUSERMESSAGE = _descriptor.Descriptor(
    name="GetUserMessage",
    full_name="syft.grid.messages.GetUserMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.grid.messages.GetUserMessage.msg_id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.grid.messages.GetUserMessage.address",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="content",
            full_name="syft.grid.messages.GetUserMessage.content",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="reply_to",
            full_name="syft.grid.messages.GetUserMessage.reply_to",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=431,
    serialized_end=584,
)


_GETUSERRESPONSE = _descriptor.Descriptor(
    name="GetUserResponse",
    full_name="syft.grid.messages.GetUserResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.grid.messages.GetUserResponse.msg_id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="status_code",
            full_name="syft.grid.messages.GetUserResponse.status_code",
            index=1,
            number=2,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="content",
            full_name="syft.grid.messages.GetUserResponse.content",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.grid.messages.GetUserResponse.address",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=587,
    serialized_end=721,
)


_GETUSERSMESSAGE = _descriptor.Descriptor(
    name="GetUsersMessage",
    full_name="syft.grid.messages.GetUsersMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.grid.messages.GetUsersMessage.msg_id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.grid.messages.GetUsersMessage.address",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="content",
            full_name="syft.grid.messages.GetUsersMessage.content",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="reply_to",
            full_name="syft.grid.messages.GetUsersMessage.reply_to",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=724,
    serialized_end=878,
)


_GETUSERSRESPONSE = _descriptor.Descriptor(
    name="GetUsersResponse",
    full_name="syft.grid.messages.GetUsersResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.grid.messages.GetUsersResponse.msg_id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="status_code",
            full_name="syft.grid.messages.GetUsersResponse.status_code",
            index=1,
            number=2,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="content",
            full_name="syft.grid.messages.GetUsersResponse.content",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.grid.messages.GetUsersResponse.address",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=881,
    serialized_end=1016,
)


_DELETEUSERMESSAGE = _descriptor.Descriptor(
    name="DeleteUserMessage",
    full_name="syft.grid.messages.DeleteUserMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.grid.messages.DeleteUserMessage.msg_id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.grid.messages.DeleteUserMessage.address",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="content",
            full_name="syft.grid.messages.DeleteUserMessage.content",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="reply_to",
            full_name="syft.grid.messages.DeleteUserMessage.reply_to",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1019,
    serialized_end=1175,
)


_DELETEUSERRESPONSE = _descriptor.Descriptor(
    name="DeleteUserResponse",
    full_name="syft.grid.messages.DeleteUserResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.grid.messages.DeleteUserResponse.msg_id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="status_code",
            full_name="syft.grid.messages.DeleteUserResponse.status_code",
            index=1,
            number=2,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="content",
            full_name="syft.grid.messages.DeleteUserResponse.content",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.grid.messages.DeleteUserResponse.address",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1178,
    serialized_end=1315,
)


_UPDATEUSERMESSAGE = _descriptor.Descriptor(
    name="UpdateUserMessage",
    full_name="syft.grid.messages.UpdateUserMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.grid.messages.UpdateUserMessage.msg_id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.grid.messages.UpdateUserMessage.address",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="content",
            full_name="syft.grid.messages.UpdateUserMessage.content",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="reply_to",
            full_name="syft.grid.messages.UpdateUserMessage.reply_to",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1318,
    serialized_end=1474,
)


_UPDATEUSERRESPONSE = _descriptor.Descriptor(
    name="UpdateUserResponse",
    full_name="syft.grid.messages.UpdateUserResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.grid.messages.UpdateUserResponse.msg_id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="status_code",
            full_name="syft.grid.messages.UpdateUserResponse.status_code",
            index=1,
            number=2,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="content",
            full_name="syft.grid.messages.UpdateUserResponse.content",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.grid.messages.UpdateUserResponse.address",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1477,
    serialized_end=1614,
)


_SEARCHUSERSMESSAGE = _descriptor.Descriptor(
    name="SearchUsersMessage",
    full_name="syft.grid.messages.SearchUsersMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.grid.messages.SearchUsersMessage.msg_id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.grid.messages.SearchUsersMessage.address",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="content",
            full_name="syft.grid.messages.SearchUsersMessage.content",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="reply_to",
            full_name="syft.grid.messages.SearchUsersMessage.reply_to",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1617,
    serialized_end=1774,
)


_SEARCHUSERSRESPONSE = _descriptor.Descriptor(
    name="SearchUsersResponse",
    full_name="syft.grid.messages.SearchUsersResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.grid.messages.SearchUsersResponse.msg_id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="status_code",
            full_name="syft.grid.messages.SearchUsersResponse.status_code",
            index=1,
            number=2,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="content",
            full_name="syft.grid.messages.SearchUsersResponse.content",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.grid.messages.SearchUsersResponse.address",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1777,
    serialized_end=1915,
)

_CREATEUSERMESSAGE.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_CREATEUSERMESSAGE.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_CREATEUSERMESSAGE.fields_by_name[
    "reply_to"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_CREATEUSERRESPONSE.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_CREATEUSERRESPONSE.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_GETUSERMESSAGE.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_GETUSERMESSAGE.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_GETUSERMESSAGE.fields_by_name[
    "reply_to"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_GETUSERRESPONSE.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_GETUSERRESPONSE.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_GETUSERSMESSAGE.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_GETUSERSMESSAGE.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_GETUSERSMESSAGE.fields_by_name[
    "reply_to"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_GETUSERSRESPONSE.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_GETUSERSRESPONSE.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_DELETEUSERMESSAGE.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_DELETEUSERMESSAGE.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_DELETEUSERMESSAGE.fields_by_name[
    "reply_to"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_DELETEUSERRESPONSE.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_DELETEUSERRESPONSE.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_UPDATEUSERMESSAGE.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_UPDATEUSERMESSAGE.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_UPDATEUSERMESSAGE.fields_by_name[
    "reply_to"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_UPDATEUSERRESPONSE.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_UPDATEUSERRESPONSE.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_SEARCHUSERSMESSAGE.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_SEARCHUSERSMESSAGE.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_SEARCHUSERSMESSAGE.fields_by_name[
    "reply_to"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_SEARCHUSERSRESPONSE.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_SEARCHUSERSRESPONSE.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
DESCRIPTOR.message_types_by_name["CreateUserMessage"] = _CREATEUSERMESSAGE
DESCRIPTOR.message_types_by_name["CreateUserResponse"] = _CREATEUSERRESPONSE
DESCRIPTOR.message_types_by_name["GetUserMessage"] = _GETUSERMESSAGE
DESCRIPTOR.message_types_by_name["GetUserResponse"] = _GETUSERRESPONSE
DESCRIPTOR.message_types_by_name["GetUsersMessage"] = _GETUSERSMESSAGE
DESCRIPTOR.message_types_by_name["GetUsersResponse"] = _GETUSERSRESPONSE
DESCRIPTOR.message_types_by_name["DeleteUserMessage"] = _DELETEUSERMESSAGE
DESCRIPTOR.message_types_by_name["DeleteUserResponse"] = _DELETEUSERRESPONSE
DESCRIPTOR.message_types_by_name["UpdateUserMessage"] = _UPDATEUSERMESSAGE
DESCRIPTOR.message_types_by_name["UpdateUserResponse"] = _UPDATEUSERRESPONSE
DESCRIPTOR.message_types_by_name["SearchUsersMessage"] = _SEARCHUSERSMESSAGE
DESCRIPTOR.message_types_by_name["SearchUsersResponse"] = _SEARCHUSERSRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

CreateUserMessage = _reflection.GeneratedProtocolMessageType(
    "CreateUserMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _CREATEUSERMESSAGE,
        "__module__": "proto.grid.messages.user_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.CreateUserMessage)
    },
)
_sym_db.RegisterMessage(CreateUserMessage)

CreateUserResponse = _reflection.GeneratedProtocolMessageType(
    "CreateUserResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _CREATEUSERRESPONSE,
        "__module__": "proto.grid.messages.user_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.CreateUserResponse)
    },
)
_sym_db.RegisterMessage(CreateUserResponse)

GetUserMessage = _reflection.GeneratedProtocolMessageType(
    "GetUserMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETUSERMESSAGE,
        "__module__": "proto.grid.messages.user_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.GetUserMessage)
    },
)
_sym_db.RegisterMessage(GetUserMessage)

GetUserResponse = _reflection.GeneratedProtocolMessageType(
    "GetUserResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETUSERRESPONSE,
        "__module__": "proto.grid.messages.user_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.GetUserResponse)
    },
)
_sym_db.RegisterMessage(GetUserResponse)

GetUsersMessage = _reflection.GeneratedProtocolMessageType(
    "GetUsersMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETUSERSMESSAGE,
        "__module__": "proto.grid.messages.user_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.GetUsersMessage)
    },
)
_sym_db.RegisterMessage(GetUsersMessage)

GetUsersResponse = _reflection.GeneratedProtocolMessageType(
    "GetUsersResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETUSERSRESPONSE,
        "__module__": "proto.grid.messages.user_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.GetUsersResponse)
    },
)
_sym_db.RegisterMessage(GetUsersResponse)

DeleteUserMessage = _reflection.GeneratedProtocolMessageType(
    "DeleteUserMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _DELETEUSERMESSAGE,
        "__module__": "proto.grid.messages.user_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.DeleteUserMessage)
    },
)
_sym_db.RegisterMessage(DeleteUserMessage)

DeleteUserResponse = _reflection.GeneratedProtocolMessageType(
    "DeleteUserResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _DELETEUSERRESPONSE,
        "__module__": "proto.grid.messages.user_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.DeleteUserResponse)
    },
)
_sym_db.RegisterMessage(DeleteUserResponse)

UpdateUserMessage = _reflection.GeneratedProtocolMessageType(
    "UpdateUserMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _UPDATEUSERMESSAGE,
        "__module__": "proto.grid.messages.user_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.UpdateUserMessage)
    },
)
_sym_db.RegisterMessage(UpdateUserMessage)

UpdateUserResponse = _reflection.GeneratedProtocolMessageType(
    "UpdateUserResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _UPDATEUSERRESPONSE,
        "__module__": "proto.grid.messages.user_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.UpdateUserResponse)
    },
)
_sym_db.RegisterMessage(UpdateUserResponse)

SearchUsersMessage = _reflection.GeneratedProtocolMessageType(
    "SearchUsersMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _SEARCHUSERSMESSAGE,
        "__module__": "proto.grid.messages.user_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.SearchUsersMessage)
    },
)
_sym_db.RegisterMessage(SearchUsersMessage)

SearchUsersResponse = _reflection.GeneratedProtocolMessageType(
    "SearchUsersResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _SEARCHUSERSRESPONSE,
        "__module__": "proto.grid.messages.user_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.SearchUsersResponse)
    },
)
_sym_db.RegisterMessage(SearchUsersResponse)


# @@protoc_insertion_point(module_scope)



# import uuid
# third party
import pytest

# syft absolute
# from syft.core.common.uid import UID
from syft.core.common.serde.serializable import Serializable
from syft.core.common.serde.serialize import _serialize

# import syft as sy


def test_object_with_no_serialize_wrapper() -> None:
    """
    Test if an object that is Serializable but does not implement the _sy_serializable_wrapper_type
    throws an exception when trying to serialize.
    """

    class TestObject(Serializable):
        pass

    with pytest.raises(Exception):
        _serialize(TestObject())



# stdlib
from typing import List as TypeList
from typing import Type as TypeType

# third party
import pytest
import torch as th

# syft absolute
import syft as sy


@pytest.mark.slow
def test_torch_function(client: sy.VirtualMachineClient) -> None:
    x = th.tensor([[-0.1, 0.1], [0.2, 0.3]])
    ptr_x = x.send(client)
    ptr_res = client.torch.zeros_like(ptr_x)
    res = ptr_res.get()

    assert (res == th.tensor([[0.0, 0.0], [0.0, 0.0]])).all()


def test_path_cache() -> None:
    short_fqn = "torch.nn.Conv2d"
    conv2d_paths = [
        "torch.nn.modules.conv.Conv2d",
        "torch.nn.modules.Conv2d",
        "torch.nn.Conv2d",
    ]

    refs: TypeList[TypeType] = []
    for path in conv2d_paths:
        klass = sy.lib_ast.query(path, obj_type=th.nn.Conv2d)
        assert klass == sy.lib_ast.torch.nn.Conv2d
        assert klass.name == "Conv2d"
        assert klass.path_and_name == short_fqn
        assert all(klass == ref for ref in refs)
        refs.append(klass)



# stdlib
from typing import Any
from typing import Dict

# syft absolute
import syft as sy
from syft import serialize
from syft.core.io.address import Address
from syft.grid.messages.dataset_messages import CreateDatasetMessage
from syft.grid.messages.dataset_messages import CreateDatasetResponse
from syft.grid.messages.dataset_messages import DeleteDatasetMessage
from syft.grid.messages.dataset_messages import DeleteDatasetResponse
from syft.grid.messages.dataset_messages import GetDatasetMessage
from syft.grid.messages.dataset_messages import GetDatasetResponse
from syft.grid.messages.dataset_messages import GetDatasetsMessage
from syft.grid.messages.dataset_messages import GetDatasetsResponse
from syft.grid.messages.dataset_messages import UpdateDatasetMessage
from syft.grid.messages.dataset_messages import UpdateDatasetResponse


def test_create_dataset_message_serde(node: sy.VirtualMachine) -> None:
    target = Address(name="Alice")

    request_content = {
        "dataset": ["<tensor_id>", "<tensor_id>", "<tensor_id>", "<tensor_id>"],
        "description": "Dataset Description",
        "tags": ["#x", "#data-sample"],
        "pointable": True,
        "read-permission": ["user-id1", "user-id2", "user-id3"],
        "write-permission": ["user-id1", "user-id5", "user-id9"],
    }
    msg = CreateDatasetMessage(
        address=target,
        content=request_content,
        reply_to=node.address,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_create_dataset_response_serde() -> None:
    target = Address(name="Alice")

    request_content = {"msg": "Dataset created succesfully!"}
    msg = CreateDatasetResponse(
        address=target,
        status_code=200,
        content=request_content,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_delete_dataset_message_serde(node: sy.VirtualMachine) -> None:
    target = Address(name="Alice")

    request_content = {"dataset_id": "f2a6as5d16fasd"}
    msg = DeleteDatasetMessage(
        address=target,
        content=request_content,
        reply_to=node.address,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_delete_dataset_response_serde() -> None:
    target = Address(name="Alice")

    content = {"msg": "Dataset deleted successfully!"}
    msg = DeleteDatasetResponse(
        status_code=200,
        address=target,
        content=content,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_update_dataset_message_serde(node: sy.VirtualMachine) -> None:
    target = Address(name="Alice")

    content = {
        "dataset": ["<tensor_id>", "<tensor_id>", "<tensor_id>", "<tensor_id>"],
        "description": "Dataset Description",
        "tags": ["#x", "#data-sample"],
        "pointable": True,
        "read-permission": ["user-id1", "user-id2", "user-id3"],
        "write-permission": ["user-id1", "user-id5", "user-id9"],
    }
    msg = UpdateDatasetMessage(
        address=target,
        content=content,
        reply_to=node.address,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_update_dataset_response_serde() -> None:
    target = Address(name="Alice")

    request_content = {"msg": "Dataset updated successfully!"}
    msg = UpdateDatasetResponse(
        address=target,
        status_code=200,
        content=request_content,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_get_dataset_message_serde(node: sy.VirtualMachine) -> None:
    target = Address(name="Alice")

    content = {"dataset_id": "eqw9e4a5d846"}
    msg = GetDatasetMessage(
        address=target,
        content=content,
        reply_to=node.address,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_get_dataset_response_serde() -> None:
    target = Address(name="Alice")

    content = {
        "dataset": ["<tensor_id>", "<tensor_id>", "<tensor_id>", "<tensor_id>"],
        "description": "Dataset Description",
        "tags": ["#x", "#data-sample"],
        "pointable": True,
        "read-permission": ["user-id1", "user-id2", "user-id3"],
        "write-permission": ["user-id1", "user-id5", "user-id9"],
    }

    msg = GetDatasetResponse(
        address=target,
        status_code=200,
        content=content,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_get_all_datasets_message_serde(node: sy.VirtualMachine) -> None:
    target = Address(name="Alice")

    content: Dict[Any, Any] = {}
    msg = GetDatasetsMessage(
        address=target,
        content=content,
        reply_to=node.address,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_get_all_datasets_response_serde() -> None:
    target = Address(name="Alice")

    request_content = {
        "workers": {
            "626sadaf631": {
                "dataset": ["<tensor_id>", "<tensor_id>", "<tensor_id>", "<tensor_id>"],
                "description": "Dataset Description",
                "tags": ["#x", "#data-sample"],
                "pointable": True,
                "read-permission": ["user-id1", "user-id2", "user-id3"],
                "write-permission": ["user-id1", "user-id5", "user-id9"],
            },
            "a84ew64wq6e": {
                "dataset": ["<tensor_id>", "<tensor_id>", "<tensor_id>", "<tensor_id>"],
                "description": "Dataset Description",
                "tags": ["#x", "#data-sample"],
                "pointable": False,
                "read-permission": ["user-id1", "user-id2", "user-id3"],
                "write-permission": [],
            },
        }
    }

    msg = GetDatasetsResponse(
        address=target,
        status_code=200,
        content=request_content,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2



# stdlib
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Type

# third party
from pandas import DataFrame

# syft relative
from ....core.common.message import SyftMessage


class GridRequestAPI:
    def __init__(
        self,
        send: Callable,
        create_msg: Optional[Type[SyftMessage]] = None,
        get_msg: Optional[Type[SyftMessage]] = None,
        get_all_msg: Optional[Type[SyftMessage]] = None,
        update_msg: Optional[Type[SyftMessage]] = None,
        delete_msg: Optional[Type[SyftMessage]] = None,
        response_key: str = "",
    ):
        self.__create_message = create_msg
        self.__get_message = get_msg
        self.__get_all_message = get_all_msg
        self.__update_message = update_msg
        self.__delete_message = delete_msg
        self.__send = send
        self.__response_key = response_key

    @property
    def send(self) -> Callable:
        return self.__send

    def create(self, **kwargs: Any) -> Dict[str, str]:
        return self.__send(grid_msg=self.__create_message, content=kwargs)

    def get(self, **kwargs: Any) -> Any:
        return self.to_obj(self.__send(grid_msg=self.__get_message, content=kwargs))

    def all(self, pandas: bool = False) -> Dict[str, Any]:
        result = self.__send(grid_msg=self.__get_all_message)
        if pandas:
            result = DataFrame(result)

        return result

    def update(self, **kwargs: Any) -> Dict[str, Any]:
        return self.__send(grid_msg=self.__update_message, content=kwargs)

    def delete(self, **kwargs: Any) -> Dict[str, Any]:
        return self.__send(grid_msg=self.__delete_message, content=kwargs)

    def to_obj(self, result: Any) -> Any:
        if result:
            _class_name = self.__response_key.capitalize()
            result = type(_class_name, (object,), result)()

        return result



# stdlib
import ast
from collections import OrderedDict
import copy
import os
from pathlib import Path
import sys
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

# third party
import torch

# syft absolute
import syft as sy

# syft relative
from ...generate_wrapper import GenerateWrapper

# from ...core.pointer.pointer import Pointer
from ...lib.util import full_name_with_qualname
from ...logger import critical
from ...logger import info
from ...logger import traceback_and_raise
from ...proto.lib.torch.module_pb2 import Module as Module_PB
from ..python.collections import OrderedDict as SyOrderedDict

# from ...core.node.common.service.auth import AuthorizationException


def repr_to_kwargs(repr_str: str) -> Tuple[List[Any], Dict[Any, Any]]:
    # for example: repr_str = Conv2d(...).extra_repr()
    # produces:    > str("1, 32, kernel_size=(3, 3), stride=(1, 1)")
    # then we just have to split it into args and kwargs
    # using ast.literal_eval we can use python to give us the real primitive types
    # from the strings in a safe way
    # str("1 ") becomes int(1)
    # str("(1, 2) ") becomes tuple(1, 2)
    args: List[Any] = []
    kwargs: Dict[Any, Any] = {}
    parts = repr_str.split(",")

    # tuples are split by commas as well, so we will keep a tab on open parentheses
    # then concat with "," until we find a close parentheses
    # TODO: make work nested with a count and add tests
    para_open = False
    buffer = ""
    for part in parts:
        try:
            if "(" in part:
                para_open = True
                buffer = ""
            if para_open is True:
                buffer += part + ","
                if ")" in part:
                    # remove trailing ,
                    part = buffer[:-1]
                    buffer = ""
                    para_open = False
                else:
                    continue

            string = part.strip()
            if "=" not in string:
                # its an arg
                arg = ast.literal_eval(string)
                args.append(arg)
            else:
                # its a kwarg
                kv = string.split("=")
                key = str(kv[0])
                string = kv[1].strip()
                value = ast.literal_eval(string)
                kwargs[key.strip()] = value
        except Exception as e:
            info(f"ast.literal_eval failed to parse part: {string}. {e}")

    return (args, kwargs)


class Module:
    """
    This is our equivalent of torch.nn.Module and aims to have the same external
    interface. We will need to support both torch Modules and Module Pointers.
    """

    def __init__(self, torch_ref: Any) -> None:
        self.setup(torch_ref=torch_ref)

    def setup(self, torch_ref: Any) -> None:
        # the remote torch means the model is remote
        self.remote_model: Optional["Module"] = None
        self.local_model: Optional["Module"] = None
        self.duet = None
        if "syft" in full_name_with_qualname(klass=type(torch_ref)):
            info("> Creating remote model")
            self.is_local = False
        else:
            # otherwise we have a local model
            info("> Creating local model")
            self.is_local = True

        self.torch_ref = torch_ref
        self.training = False
        self._modules: OrderedDict[str, Module] = OrderedDict()
        real_module = torch_ref.nn.Module()
        self.__dict__["real_module"] = real_module  # bypass getattr/setattr
        # if issubclass(type(real_module), Pointer):
        #     try:
        #         # TODO: this needs fixing but should be on by default for now
        #         # https://github.com/OpenMined/PySyft/issues/5242
        #         real_module.searchable = True
        #     except AuthorizationException as e:
        #         print(f"Cant make real_module searchable. {e}")

    def __setattr__(self, name: str, value: Union[Any, "Module"]) -> None:
        # this is how we catch the modules being set during subclass init
        # bug where torch.nn.modules isn't the full name on some imports
        # TODO: fix this properly
        if "torch.nn" in full_name_with_qualname(klass=type(value)):
            modules = self.__dict__.get("_modules")
            if modules is not None:
                modules[name] = value

            # attach all the sub modules to a real module so that we can have a
            # remote module pointer that acts like a real model
            real_module: Optional[OrderedDict] = self.__dict__.get("real_module")
            if real_module is not None:
                real_module.add_module(name, value)  # type: ignore
        else:
            object.__setattr__(self, name, value)

    def __getattr__(self, name: str) -> Union[Any, "Module"]:
        modules: Optional[OrderedDict] = self.__dict__.get("_modules")
        if modules is not None:
            if name in modules:
                return modules[name]

        return object.__getattribute__(self, name)

    def train(self, mode: bool = True) -> "Module":
        self.training = mode
        for _, module in self.modules.items():
            module.train(mode)
        return self

    def eval(self) -> "Module":
        return self.train(False)

    def __call__(
        self, *args: Union[List[Any], Tuple[Any, ...]], **kwargs: Dict[Any, Any]
    ) -> Any:
        return self.forward(*args, **kwargs)

    @property
    def modules(self) -> OrderedDict:
        modules = self.__dict__.get("_modules")
        if modules is not None:
            return modules
        return OrderedDict()

    # local list of remote ListPointers of TensorPointers
    def parameters(self, recurse: bool = True) -> Optional[List[Any]]:
        params_list: Optional[List[Any]] = None

        if self.is_local is True:
            # we are local so use normal torch params
            params_list = []
        for _, module in self.modules.items():
            params = module.parameters(recurse)
            if params_list is None:
                # only on remote create a remote list so we can concat the param list
                # pointers without having to actually get them
                self.duet = params.client
                params_list = self.duet.syft.lib.python.List()  # type: ignore
            # either way lets concat them until we have a big list of parameters
            params_list += params
        return params_list

    def cuda(self, device: Any) -> "Module":
        for _, module in self.modules.items():
            module.cuda(device)
        return self

    def cpu(self) -> "Module":
        for _, module in self.modules.items():
            module.cpu()
        return self

    def load_state_dict(self, input: Union[str, os.PathLike, Dict[str, Any]]) -> None:
        if not self.is_local:
            info("> This model is remote so try calling .get()")
            return None

        state_dict = {}
        if isinstance(input, (str, os.PathLike)):
            with open(Path(input), "rb") as f:
                state_dict = torch.load(f)
        else:
            state_dict = dict(input)

        if not issubclass(type(state_dict), dict):
            traceback_and_raise(
                f"  Invalid input: {type(input)}. "
                + "Try inputting a state_dict or .pth file."
            )

        info("> Loading model weights")
        layers: Dict[str, Any] = {}
        for save_key, values in state_dict.items():
            parts = save_key.split(".")
            if len(parts) < 2:
                info(f"  state dict key is too short: {save_key}")
                continue
            layer = parts[0]
            attr = parts[1]
            if layer not in layers:
                layers[layer] = {}
            layers[layer][attr] = values

        for layer, sd in layers.items():
            local_layer = getattr(self, layer, None)
            if local_layer is not None and hasattr(local_layer, "load_state_dict"):
                d = local_layer.load_state_dict(sd)
                info(f"  {layer} state dict loaded with: {d}")
            else:
                info(f"  Model doesnt have layer {layer}")

        info("> Finished loading weights")
        return None

    def state_dict(self) -> Optional[Dict[str, Any]]:
        if not self.is_local:
            info("> This model is remote so try calling .get()")
            return None

        info("> Saving model weights")
        model_state_dict = OrderedDict()
        for name, module in self.modules.items():
            if hasattr(module, "state_dict"):
                for k, v in module.state_dict().items():
                    save_key = f"{name}.{k}"
                    model_state_dict[save_key] = v

        info("> Finished saving weights")
        return model_state_dict

    def save(self, path: Union[str, bytes, os.PathLike]) -> None:
        if not self.is_local:
            info("> This model is remote so try calling .get()")
            return

        state_dict = self.state_dict()
        torch.save(state_dict, path)

    def load(self, path: Union[str, os.PathLike]) -> None:
        if not self.is_local:
            info("> This model is remote so try calling .get()")
            return

        self.load_state_dict(input=path)

    def send(self, client: Any, send_parameters: bool = True) -> Any:
        if not self.is_local:
            info("> This model is remote so try calling .get()")
            return

        info("> Sending local model")

        remote_model = copy.copy(self)
        remote_model.setup(torch_ref=client.torch)
        remote_model.duet = client

        for name, module in self.modules.items():
            fqn = full_name_with_qualname(klass=type(module))
            klass = client.lib_ast.query(fqn, obj_type=type(module))
            module_repr = module.extra_repr()
            args, kwargs = repr_to_kwargs(repr_str=module_repr)
            remote_module_ptr = klass(*args, **kwargs)
            remote_model.__setattr__(name, remote_module_ptr)

            # if the remote module has state_dict lets get it
            if (
                send_parameters
                and hasattr(module, "state_dict")
                and hasattr(remote_module_ptr, "load_state_dict")
            ):
                local_state_ord_dict = module.state_dict()
                # cast to dict because OrderedDict is not supported

                # get a blocking copy of the state_dict
                info(f"  Sending local layer: {name}")
                # cant import Dict / PrimitiveFactory due to circular imports
                remote_state_dict_ptr = client.syft.lib.python.Dict(
                    dict(local_state_ord_dict)
                )
                # iterate through the key, values
                # weights and biases should be in there
                remote_module_ptr.load_state_dict(remote_state_dict_ptr)

        info("\n> Finished sending local model <\n\n")
        self.remote_model = remote_model
        return self.remote_model

    def get(
        self,
        request_block: bool = False,
        timeout_secs: int = 20,
        reason: str = "",
        delete_obj: bool = False,
    ) -> Optional["Module"]:

        if self.is_local:
            info("> This model is local. Maybe you meant to call .send()?")
            return None

        info("> Downloading remote model")

        local_model = copy.copy(self)
        local_model.setup(torch_ref=torch)
        local_model.duet = self.duet

        for layer_name, module in self.modules.items():
            module_parts = module.path_and_name.split(".")
            klass_name = module_parts.pop()
            klass = getattr(sys.modules[".".join(module_parts)], klass_name)
            repr_ptr = module.extra_repr()

            module_repr = repr_ptr.get(
                request_block=request_block,
                reason=reason,
                timeout_secs=timeout_secs,
            )

            if module_repr is None:
                info(f"  Request for {reason} extra_repr failed, skipping layer")
                continue

            args, kwargs = repr_to_kwargs(repr_str=module_repr.upcast())
            local_module = klass(*args, **kwargs)

            # the local real module has been set on the sy module
            local_model.__setattr__(layer_name, local_module)

            try:
                # if the remote module has state_dict lets get it
                if hasattr(module, "state_dict") and hasattr(
                    local_module, "load_state_dict"
                ):
                    info("loading remote state dict")
                    sd_ptr = module.state_dict()
                    # get a blocking copy of the state_dict
                    info(f"  Downloading remote layer: {layer_name}")
                    state_dict = sd_ptr.get(
                        request_block=request_block,
                        reason=reason,
                        timeout_secs=timeout_secs,
                        delete_obj=delete_obj,
                    )
                    # We have to recreate the OrderedDict for load_state_dict to work
                    ordered_state_dict = OrderedDict()
                    for elem, item in state_dict.items():
                        ordered_state_dict[str(elem)] = item
                    # iterate through the key, values
                    # weights and biases should be in there
                    if state_dict is not None:
                        # TODO: support torch.nn.modules.module._IncompatibleKeys
                        local_module.load_state_dict(ordered_state_dict)
                    else:
                        info(
                            f"  Failed to get {layer_name} state_dict, skipping layer."
                        )

            except Exception as e:
                critical(f"  Failed to download remote state for {layer_name}.")
                traceback_and_raise(e)

        info("\n> Finished downloading remote model <\n\n")
        self.local_model = local_model
        return self.local_model

    # zero them so we know they are copied
    def zero_layers(self) -> None:
        for m in self.modules.values():
            if hasattr(m, "weight"):
                m.weight.requires_grad_(False).zero_()
            if hasattr(m, "bias"):
                m.bias.requires_grad_(False).zero_()

    # easy way to check the weights have changed
    def debug_sum_layers(self) -> None:
        info("> Summing layers for debugging: ")
        for n, m in self.modules.items():
            if hasattr(m, "state_dict"):
                if self.is_local:
                    state_dict = m.state_dict()
                else:
                    state_dict = m.state_dict().get()

                for k, v in state_dict.items():
                    if hasattr(v, "sum"):
                        s = v.sum().item()
                        info(f"  Layer {n} sum({k}): {s}")


def object2proto(obj: torch.nn.Module, is_child: bool = False) -> Module_PB:
    proto = Module_PB()
    if "torch.nn." in type(obj).__module__:
        proto.module_type = type(obj).__name__
    else:
        proto.module_type = f"_USER_DEFINED_MODULE_{type(obj).__name__}"

    proto.module_repr = obj.extra_repr()

    if not is_child:
        proto.state_dict.CopyFrom(sy.serialize(SyOrderedDict(obj.state_dict())))

    for n, m in obj.named_children():
        child_proto = object2proto(m, is_child=True)
        child_proto.module_name = n
        proto.children.append(child_proto)

    return proto


def proto2object(proto: Module_PB) -> torch.nn.Module:
    is_userdefined = proto.module_type.startswith("_USER_DEFINED_MODULE_")

    if is_userdefined:
        obj_type = type(
            proto.module_type.replace("_USER_DEFINED_MODULE_", ""),
            (torch.nn.Module,),
            {},
        )
    else:
        obj_type = getattr(torch.nn, proto.module_type)

    args, kwargs = repr_to_kwargs(repr_str=proto.module_repr)
    obj = obj_type(*args, **kwargs)

    for child_proto in proto.children:
        obj.add_module(child_proto.module_name, sy.deserialize(child_proto))

    if proto.state_dict.ByteSize() > 0:
        obj.load_state_dict(sy.deserialize(proto.state_dict))

    return obj


GenerateWrapper(
    wrapped_type=torch.nn.Module,
    import_path="torch.nn.Module",
    protobuf_scheme=Module_PB,
    type_object2proto=object2proto,
    type_proto2object=proto2object,
)
